#include <string.h>

#include <algorithm>
#include <vector>

#include <zephyr/drivers/uart.h>

#include "KMessagePassing.h"
#include "PAL.h"
#include "Shell.h"
#include "UART.h"
#include "Utl.h"

using namespace std;


////////////////////////////////////////////////////////////////////////////////
// UART Init state
////////////////////////////////////////////////////////////////////////////////

static const device *DEV_UART_0 = DEVICE_DT_GET(DT_NODELABEL(uart0));
static const device *DEV_UART_1 = DEVICE_DT_GET(DT_NODELABEL(uart1));
static const device *DEV_UART_USB = DEVICE_DT_GET(DT_NODELABEL(cdc_acm0));


////////////////////////////////////////////////////////////////////////////////
// UART Output state
////////////////////////////////////////////////////////////////////////////////

static const uint16_t UART_OUTPUT_PIPE_SIZE = 1000;
static KMessagePipe<char, UART_OUTPUT_PIPE_SIZE> UART_0_OUTPUT_PIPE;
static KMessagePipe<char, UART_OUTPUT_PIPE_SIZE> UART_1_OUTPUT_PIPE;
static KMessagePipe<char, UART_OUTPUT_PIPE_SIZE> UART_USB_OUTPUT_PIPE;


////////////////////////////////////////////////////////////////////////////////
// UART Input state
////////////////////////////////////////////////////////////////////////////////

class DataStreamDistributor
{
public:
    DataStreamDistributor(UART uart)
    : uart_(uart)
    {
        // Reserve id of 0 for special purposes
        idMaker_.GetNextId();   // 0
    }

    pair<bool, uint8_t> AddDataStreamCallback(function<void(const vector<uint8_t> &data)> cbFn)
    {
        auto [ok, id] = idMaker_.GetNextId();

        if (ok)
        {
            id__cbFn_[id] = cbFn;
        }

        return { ok, id };
    }

    bool SetDataStreamCallback(uint8_t id, function<void(const vector<uint8_t> &data)> cbFn)
    {
        bool retVal = false;

        if (id__cbFn_.contains(id) || id == 0)
        {
            id__cbFn_[id] = cbFn;

            retVal = true;
        }

        return retVal;
    }

    bool RemoveDataStreamCallback(uint8_t id)
    {
        idMaker_.ReturnId(id);

        return id__cbFn_.erase(id);
    }

    void Distribute(const vector<uint8_t> &data)
    {
        // distribute
        for (auto &[id, cbFn] : id__cbFn_)
        {
            // default to writing back to the uart that sent the data
            // UartTarget target(uart_);    // ehh, let's keep this functionality very simple

            // fire callback
            cbFn(data);
        }
    }

    uint8_t GetCallbackCount()
    {
        return id__cbFn_.size();
    }

private:

    UART uart_;
    IDMaker<32> idMaker_;
    unordered_map<uint8_t, function<void(const vector<uint8_t> &data)>> id__cbFn_;
};

class LineStreamDistributor
{
public:
    LineStreamDistributor(UART uart, uint16_t maxLineLen)
    : uart_(uart)
    , maxLineLen_(maxLineLen)
    {
        // Reserve id of 0 for special purposes
        idMaker_.GetNextId();   // 0
    }

    pair<bool, uint8_t> AddLineStreamCallback(function<void(const string &line)> cbFn)
    {
        auto [ok, id] = idMaker_.GetNextId();

        if (ok)
        {
            id__cbFn_[id] = cbFn;
        }

        return { ok, id };
    }

    bool SetLineStreamCallback(uint8_t id, function<void(const string &line)> cbFn)
    {
        bool retVal = false;

        if (id__cbFn_.contains(id) || id == 0)
        {
            id__cbFn_[id] = cbFn;

            retVal = true;
        }

        return retVal;
    }

    bool RemoveLineStreamCallback(uint8_t id)
    {
        idMaker_.ReturnId(id);
        
        return id__cbFn_.erase(id);
    }

    void AddData(const vector<uint8_t> &data)
    {
        if (id__cbFn_.empty()) { return; }

        for (int i = 0; i < (int)data.size(); ++i)
        {
            char c = data[i];

            if (c == '\n' || inputStream_.size() == maxLineLen_)
            {
                // remember if max line len hit
                bool wasMaxLine = inputStream_.size() == maxLineLen_;

                if (inputStream_.size())
                {
                    // make a copy (move) of the input stream on the way in
                    Evm::QueueLowPriorityWork("INPUT_STREAM", [this, inputStream = move(inputStream_)](){
                        // distribute
                        for (auto &[id, cbFn] : id__cbFn_)
                        {
                            // default to writing back to the uart that sent the data
                            UartTarget target(uart_);

                            // fire callback
                            cbFn(inputStream);
                        }
                    });

                    // clear line cache
                    inputStream_.clear();

                    // wind forward to next newline if needed
                    if (wasMaxLine)
                    {
                        for (; i < (int)data.size(); ++i)
                        {
                            if (data[i] == '\n')
                            {
                                break;
                            }
                        }
                    }
                }
            }
            else if ((isprint(c) || c == ' ' || c == '\t'))
            {
                inputStream_.push_back(c);
            }
        }
    }

    uint8_t GetCallbackCount()
    {
        return id__cbFn_.size();
    }

    void Clear()
    {
        inputStream_.clear();
    }

private:
    UART uart_;
    uint16_t maxLineLen_;
    IDMaker<32> idMaker_;
    unordered_map<uint8_t, function<void(const string &)>> id__cbFn_;
    string inputStream_;
};


static const uint16_t UART_INPUT_PIPE_SIZE = 1002;
static const uint16_t UART_INPUT_MAX_LINE_LEN = 1000;

static DataStreamDistributor UART_1_INPUT_DATA_STREAM_DISTRIBUTOR(UART::UART_1);
static vector<uint8_t> UART_1_INPUT_VEC;
static LineStreamDistributor UART_1_INPUT_LINE_STREAM_DISTRIBUTOR(UART::UART_1, UART_INPUT_MAX_LINE_LEN);
static KSemaphore UART_1_INPUT_SEM;

static DataStreamDistributor UART_USB_INPUT_DATA_STREAM_DISTRIBUTOR(UART::UART_USB);
static vector<uint8_t> UART_USB_INPUT_VEC;
static LineStreamDistributor UART_USB_INPUT_LINE_STREAM_DISTRIBUTOR(UART::UART_USB, UART_INPUT_MAX_LINE_LEN);
static KSemaphore UART_USB_INPUT_SEM;

static const uint8_t READ_BUF_SIZE = 50;
static uint8_t READ_BUF[READ_BUF_SIZE];


////////////////////////////////////////////////////////////////////////////////
// UART Output
////////////////////////////////////////////////////////////////////////////////

vector<UART> uartStack = { UART::UART_0 };

static const device *UartGetDevice()
{
    const device *dev = DEV_UART_0;

    UART uart = uartStack.back();

    if      (uart == UART::UART_1)   { dev = DEV_UART_1; }
    else if (uart == UART::UART_USB) { dev = DEV_UART_USB; }
    
    return dev;
}

static KMessagePipe<char, UART_OUTPUT_PIPE_SIZE> &UartGetPipe()
{
    KMessagePipe<char, UART_OUTPUT_PIPE_SIZE> *ptr = &UART_0_OUTPUT_PIPE;

    UART uart = uartStack.back();

    if      (uart == UART::UART_1)   { ptr = &UART_1_OUTPUT_PIPE; }
    else if (uart == UART::UART_USB) { ptr = &UART_USB_OUTPUT_PIPE; }
    
    return *ptr;
}



////////////////////////////////////////////////////////////////////////////////
// UART Input Handling
////////////////////////////////////////////////////////////////////////////////

static void UartInputIrqHandler(const struct device *dev, void * /* user_data */)
{
    KSemaphore      *sem = nullptr;
    vector<uint8_t> *vec = nullptr;
    DataStreamDistributor *dist = nullptr;

    if (dev == DEV_UART_1)
    {
        sem = &UART_1_INPUT_SEM;
        vec = &UART_1_INPUT_VEC;
        dist = &UART_1_INPUT_DATA_STREAM_DISTRIBUTOR;
    }
    else if (dev == DEV_UART_USB)
    {
        sem = &UART_USB_INPUT_SEM;
        vec = &UART_USB_INPUT_VEC;
        dist = &UART_USB_INPUT_DATA_STREAM_DISTRIBUTOR;
    }

    while (uart_irq_update(dev) && uart_irq_is_pending(dev))
    {
        // "uart_fifo_read must be called until it reads all available data in the FIFO
        // (i.e. until it returns less data than was requested)."
        if (uart_irq_rx_ready(dev))
        {
            int recvLen = uart_fifo_read(dev, READ_BUF, READ_BUF_SIZE);

            // if you got data, and there's a place to put it and someone wants it
            if (dist->GetCallbackCount() == 0)
            {
                vec->clear();
            }
            else if (recvLen > 0 && vec->size() < UART_INPUT_PIPE_SIZE)
            {
                // calculate how many bytes will fit
                int bytesToCopy = min((size_t)UART_INPUT_PIPE_SIZE - vec->size(), (size_t)recvLen);

                for (int i = 0; i < bytesToCopy; ++i)
                {
                    char c = (char)READ_BUF[i];
                    vec->push_back(c);
                }

                sem->Give();
            }
        }
    }
}


////////////////////////////////////////////////////////////////////////////////
// Threads dedicated to cranking out uart at low priority
// 
// https://docs.zephyrproject.org/3.0.0/reference/peripherals/uart.html
//
// It is wasteful to have so much stack space for all the threads, could
// consolidate down to one.  Would require:
// - use k_poll
//   - change pipe to be fifo (only fifo is compatible, not pipe)
// - change uart_poll_out to use interrupts
//   - as seen in cdc_acm example
//
// For now this is "fine" since I'm sick of dealing with it.
//
////////////////////////////////////////////////////////////////////////////////

#define STACKSIZE 1024
#define PRIORITY 7

//
// UART 0
// (input handled by shell)
//
static void ThreadFnUART0Output()
{
    while (true)
    {
        char c;
        if (UART_0_OUTPUT_PIPE.Get(c, K_FOREVER))
        {
            uart_poll_out(DEV_UART_0, c);
        }
    }
}
K_THREAD_DEFINE(ThreadUART0Output, STACKSIZE, ThreadFnUART0Output, NULL, NULL, NULL, PRIORITY, 0, 0);

//
// UART 1
//
static void ThreadFnUART1Input()
{
    while (true)
    {
        if (UART_1_INPUT_SEM.Take(K_FOREVER))
        {
            // lock out interrupts so we can safely manipulate the captured data vector
            IrqLock lock;
            PAL.SchedulerLock();

            // pass all the data to the line stream
            UART_1_INPUT_DATA_STREAM_DISTRIBUTOR.Distribute(UART_1_INPUT_VEC);

            // erase all data
            UART_1_INPUT_VEC.clear();
            PAL.SchedulerUnlock();

            // reset the semaphore, all data is drained by now
            UART_1_INPUT_SEM.Reset();
        }
    }
}
K_THREAD_DEFINE(ThreadUART1Input, STACKSIZE, ThreadFnUART1Input, NULL, NULL, NULL, PRIORITY, 0, 0);

static void ThreadFnUART1Output()
{
    while (true)
    {
        char c;
        if (UART_1_OUTPUT_PIPE.Get(c, K_FOREVER))
        {
            uart_poll_out(DEV_UART_1, c);
        }
    }
}
K_THREAD_DEFINE(ThreadUART1Output, STACKSIZE, ThreadFnUART1Output, NULL, NULL, NULL, PRIORITY, 0, 0);


//
// UART USB
//
static void ThreadFnUARTUSBOutput()
{
    bool skippedLast = false;
    while (true)
    {
        char c;
        if (UART_USB_OUTPUT_PIPE.Get(c, K_FOREVER))
        {
            // consume the bytes, but only attempt to output if USB serial connected
            uint32_t dtr;
            uart_line_ctrl_get(DEV_UART_USB, UART_LINE_CTRL_DTR, &dtr);
            if (dtr)
            {
                // sad kludge to protect against:
                // - sending a stream of data to an endpoint which disconnects midway through
                // - unknowingly queuing a few additional bytes in the USB output buffer
                // - those bytes lingering and later getting sent on a new connection, corrupting the stream
                //
                // this approach hopes:
                // - the other side is line-oriented and simply discards the malformed message
                // - too few bytes get in the buffer to form a complete message
                if (skippedLast)
                {
                    skippedLast = false;

                    for (int i = 0; i < 6; ++i)
                    {
                        uart_poll_out(DEV_UART_USB, '\n');
                    }
                }
                uart_poll_out(DEV_UART_USB, c);
            }
            else
            {
                skippedLast = true;
            }
        }
    }
}
K_THREAD_DEFINE(ThreadUARTUSBOutput, STACKSIZE, ThreadFnUARTUSBOutput, NULL, NULL, NULL, PRIORITY, 0, 0);

static void ThreadFnUARTUSBInput()
{
    while (true)
    {
        if (UART_USB_INPUT_SEM.Take(K_FOREVER))
        {
            // lock out interrupts so we can safely manipulate the captured data vector
            IrqLock lock;
            PAL.SchedulerLock();

            // pass all the data to the line stream
            UART_USB_INPUT_DATA_STREAM_DISTRIBUTOR.Distribute(UART_USB_INPUT_VEC);

            // erase all data
            UART_USB_INPUT_VEC.clear();
            PAL.SchedulerUnlock();

            // reset the semaphore, all data is drained by now
            UART_USB_INPUT_SEM.Reset();
        }
    }
}
K_THREAD_DEFINE(ThreadUARTUSBInput, STACKSIZE, ThreadFnUARTUSBInput, NULL, NULL, NULL, PRIORITY, 0, 0);


////////////////////////////////////////////////////////////////////////////////
// UART External API
////////////////////////////////////////////////////////////////////////////////

void UartPush(UART uart)
{
    uartStack.push_back(uart);
}

void UartPop()
{
    if (uartStack.size() != 1)
    {
        uartStack.pop_back();
    }
}

UART UartCurrent()
{
    return uartStack.back();
}

void UartOut(char c)
{
    uart_poll_out(UartGetDevice(), c);
}

void UartOut(const char *str)
{
    for (size_t i = 0; i < strlen(str); ++i)
    {
        UartOut(str[i]);
    }
}

void UartSend(const uint8_t *buf, uint16_t bufLen)
{
    if (PAL.InIsr())
    {
        for (int i = 0; i < bufLen; ++i)
        {
            UartOut(buf[i]);
        }
    }
    else
    {
        if (bufLen <= UART_OUTPUT_PIPE_SIZE)
        {
            PAL.SchedulerLock();

            // Make sure the output isn't disabled before sending
            bool sendOk = true;

            auto *pipe = &UartGetPipe();
            if (pipe == &UART_1_OUTPUT_PIPE)
            {
                sendOk = UartIsEnabled(UART::UART_1);
            }

            if (sendOk)
            {
                pipe->Put((char *)buf, bufLen);
            }

            PAL.SchedulerUnlock();
        }
        else
        {
            // I guess it's not going then
        }
    }
}

void UartSend(const vector<uint8_t> &byteList)
{
    UartSend(byteList.data(), byteList.size());
}

pair<bool, uint8_t> UartAddDataStreamCallback(UART uart, function<void(const vector<uint8_t> &data)> cbFn)
{
    bool retValOk = false;
    uint8_t retValId = 0;

    DataStreamDistributor *dist = nullptr;
    if (uart == UART::UART_1)
    {
        dist = &UART_1_INPUT_DATA_STREAM_DISTRIBUTOR;
    }
    else if (uart == UART::UART_USB)
    {
        dist = &UART_USB_INPUT_DATA_STREAM_DISTRIBUTOR;
    }

    if (dist)
    {
        auto [ok, id] = dist->AddDataStreamCallback(cbFn);

        retValOk = ok;
        retValId = id;
    }

    return { retValOk, retValId };
}

bool UartSetDataStreamCallback(UART uart, function<void(const vector<uint8_t> &data)> cbFn, uint8_t id)
{
    bool retVal = false;

    DataStreamDistributor *dist = nullptr;
    if (uart == UART::UART_1)
    {
        dist = &UART_1_INPUT_DATA_STREAM_DISTRIBUTOR;
    }
    else if (uart == UART::UART_USB)
    {
        dist = &UART_USB_INPUT_DATA_STREAM_DISTRIBUTOR;
    }

    if (dist)
    {
        retVal = dist->SetDataStreamCallback(id, cbFn);
    }

    return retVal;
}

bool UartRemoveDataStreamCallback(UART uart, uint8_t id)
{
    bool retVal = false;

    DataStreamDistributor *dist = nullptr;
    if (uart == UART::UART_1)
    {
        dist = &UART_1_INPUT_DATA_STREAM_DISTRIBUTOR;
    }
    else if (uart == UART::UART_USB)
    {
        dist = &UART_USB_INPUT_DATA_STREAM_DISTRIBUTOR;
    }

    if (dist)
    {
        retVal = dist->RemoveDataStreamCallback(id);
    }

    return retVal;
}

pair<bool, uint8_t> UartAddLineStreamCallback(UART uart, function<void(const string &line)> cbFn)
{
    bool retValOk = false;
    uint8_t retValId = 0;

    LineStreamDistributor *dist = nullptr;
    if (uart == UART::UART_1)
    {
        dist = &UART_1_INPUT_LINE_STREAM_DISTRIBUTOR;
    }
    else if (uart == UART::UART_USB)
    {
        dist = &UART_USB_INPUT_LINE_STREAM_DISTRIBUTOR;
    }

    if (dist)
    {
        auto [ok, id] = dist->AddLineStreamCallback(cbFn);

        retValOk = ok;
        retValId = id;
    }

    return { retValOk, retValId };
}

bool UartSetLineStreamCallback(UART uart, function<void(const string &line)> cbFn, uint8_t id)
{
    bool retVal = false;

    LineStreamDistributor *dist = nullptr;
    if (uart == UART::UART_1)
    {
        dist = &UART_1_INPUT_LINE_STREAM_DISTRIBUTOR;
    }
    else if (uart == UART::UART_USB)
    {
        dist = &UART_USB_INPUT_LINE_STREAM_DISTRIBUTOR;
    }

    if (dist)
    {
        retVal = dist->SetLineStreamCallback(id, cbFn);
    }

    return retVal;
}

bool UartRemoveLineStreamCallback(UART uart, uint8_t id)
{
    bool retVal = false;

    LineStreamDistributor *dist = nullptr;
    if (uart == UART::UART_1)
    {
        dist = &UART_1_INPUT_LINE_STREAM_DISTRIBUTOR;
    }
    else if (uart == UART::UART_USB)
    {
        dist = &UART_USB_INPUT_LINE_STREAM_DISTRIBUTOR;
    }

    if (dist)
    {
        retVal = dist->RemoveLineStreamCallback(id);
    }

    return retVal;
}

void UartDisable(UART uart)
{
    // prevent ISRs from firing, or calling this code
    IrqLock lock;

    if (uart == UART::UART_1)
    {
        NRF_UARTE1->TASKS_STOPRX = 1;
        NRF_UARTE1->TASKS_STOPTX = 1;
        
        // Wait for it to take effect
        PAL.Delay(1);
        
        // clear events again
        NRF_UARTE1->EVENTS_RXTO      = 0;
        NRF_UARTE1->EVENTS_TXSTOPPED = 0;

        // Shut off UART
        NRF_UARTE1->ENABLE = 0;

        // clobber any queued serial data inbound or outbound
        UART_1_OUTPUT_PIPE.Flush();
        UART_1_INPUT_VEC.clear();
        UART_1_INPUT_LINE_STREAM_DISTRIBUTOR.Clear();
    }
}

void UartEnable(UART uart)
{
    if (uart == UART::UART_1)
    {
        NRF_UARTE1->ENABLE = 8; // clown value
        
        // Wait for it to take effect
        PAL.Delay(1);
        
        // Enable RX/TX
        NRF_UARTE1->TASKS_STARTRX = 1;
        NRF_UARTE1->TASKS_STARTTX = 1;

        // Wait for it to take effect
        PAL.Delay(1);
    }
}

bool UartIsEnabled(UART uart)
{
    bool retVal = true;

    if (uart == UART::UART_1)
    {
        retVal = (bool)NRF_UARTE1->ENABLE;
    }

    return retVal;
}


////////////////////////////////////////////////////////////////////////////////
// Initilization
////////////////////////////////////////////////////////////////////////////////

int UartInit(const device *)
{
    Timeline::Global().Event("UartInit");

    // pre-allocate buffer so no need to do so under ISR
    UART_1_INPUT_VEC.reserve(UART_INPUT_PIPE_SIZE);
    UART_USB_INPUT_VEC.reserve(UART_INPUT_PIPE_SIZE);

    // register the line distributors with the binary distributors -- daisy chain
    UART_1_INPUT_DATA_STREAM_DISTRIBUTOR.AddDataStreamCallback([](const vector<uint8_t> &data){
        UART_1_INPUT_LINE_STREAM_DISTRIBUTOR.AddData(data);
    });
    UART_USB_INPUT_DATA_STREAM_DISTRIBUTOR.AddDataStreamCallback([](const vector<uint8_t> &data){
        UART_USB_INPUT_LINE_STREAM_DISTRIBUTOR.AddData(data);
    });

    // set up raw IRQ handling
    uart_irq_callback_set(DEV_UART_1, UartInputIrqHandler);
    uart_irq_rx_enable(DEV_UART_1);
    uart_irq_callback_set(DEV_UART_USB, UartInputIrqHandler);
    uart_irq_rx_enable(DEV_UART_USB);

    return 1;
}

int UartSetupShell(const device *)
{
    Timeline::Global().Event("UartSetupShell");
    
    // Setup shell
    Shell::AddCommand("uart.log", [](vector<string> argList){
        string uartStr = argList[0];
        UART uart = UART::UART_0;

        if      (uartStr == "1")   { uart = UART::UART_1; }
        else if (uartStr == "usb") { uart = UART::UART_USB; }

        string msg = argList[1];

        UartTarget target(uart);
        Log(msg);
    }, { .argCount = 2, .help = "UART log <uart=1> <msg>"});

    Shell::AddCommand("uart.stats", [](vector<string> argList){
        Log("UART_0");
        Log("- input queued    : N/A");
        Log("  - raw listeners : N/A");
        Log("  - line listeners: N/A");

        Log("UART_1");
        Log("- input queued    : ", UART_1_INPUT_VEC.size());
        Log("  - raw listeners : ", UART_1_INPUT_DATA_STREAM_DISTRIBUTOR.GetCallbackCount() - 1);
        Log("  - line listeners: ", UART_1_INPUT_LINE_STREAM_DISTRIBUTOR.GetCallbackCount());

        Log("UART_USB");
        Log("- input queued    : ", UART_USB_INPUT_VEC.size());
        Log("  - raw listeners : ", UART_USB_INPUT_DATA_STREAM_DISTRIBUTOR.GetCallbackCount() - 1);
        Log("  - line listeners: ", UART_USB_INPUT_LINE_STREAM_DISTRIBUTOR.GetCallbackCount());

        LogNL();
    }, { .argCount = 0, .help = "UART stats"});

    Shell::AddCommand("uart.uart1", [](vector<string> argList){
        if (argList[0] == "on") { UartEnable(UART::UART_1);  }
        else                    { UartDisable(UART::UART_1); }
    }, { .argCount = 1, .help = "UART1 <on/off>"});

    return 1;
}


#include <zephyr/init.h>
SYS_INIT(UartInit, APPLICATION, 12);
SYS_INIT(UartSetupShell, APPLICATION, 80);



